﻿using NBitcoin;
using NBitcoin.Crypto;
using NTumbleBit.BouncyCastle.Math;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Newtonsoft.Json.Converters;
using Newtonsoft.Json;
using NTumbleBit.ClassicTumbler;

namespace NTumbleBit.PuzzleSolver
{
	public enum SolverServerStates
	{
		WaitingEscrow,
		WaitingPuzzles,
		WaitingRevelation,
		WaitingBlindFactor,
		WaitingFulfillment,
		WaitingEscape,
		Completed
	}
	public class SolverServerSession : EscrowReceiver
	{
		public class SolvedPuzzle
		{
			public SolvedPuzzle()
			{

			}
			public SolvedPuzzle(PuzzleValue puzzle, SolutionKey key, PuzzleSolution solution)
			{
				Puzzle = puzzle;
				SolutionKey = key;
				Solution = solution;
			}

			public PuzzleValue Puzzle
			{
				get; set;
			}
			public SolutionKey SolutionKey
			{
				get; set;
			}
			public PuzzleSolution Solution
			{
				get; set;
			}
		}

		public new class State : EscrowReceiver.State
		{
			public SolverServerStates Status
			{
				get; set;
			}

			public SolvedPuzzle[] SolvedPuzzles
			{
				get; set;
			}
			public Key FulfillKey
			{
				get;
				set;
			}
			public Money OfferTransactionFee
			{
				get;
				set;
			}
			public TransactionSignature OfferSignature
			{
				get;
				set;
			}
			public TransactionSignature OfferClientSignature
			{
				get;
				set;
			}
			public int ETag
			{
				get;
				set;
			}

			public PubKey GetClientEscrowPubKey()
			{
				return EscrowScriptBuilder.ExtractEscrowScriptPubKeyParameters(EscrowedCoin.Redeem)
											   .EscrowKeys
											   .First(e => e != EscrowKey.PubKey);
			}
		}


		public State GetInternalState()
		{
			return Serializer.Clone(InternalState);
		}


		protected new State InternalState
		{
			get
			{
				return (State)base.InternalState;
			}
			set
			{
				base.InternalState = value;
			}
		}

		public SolverServerSession(RsaKey serverKey) : this(serverKey, null)
		{
		}

		public SolverServerSession(RsaKey serverKey, SolverParameters parameters)
		{
			parameters = parameters ?? new SolverParameters(serverKey.PubKey);
			if(serverKey == null)
				throw new ArgumentNullException(nameof(serverKey));
			if(serverKey.PubKey != parameters.ServerKey)
				throw new ArgumentNullException($"Private key not matching expected public key: {nameof(serverKey.PubKey)} != {nameof(parameters.ServerKey)}");
			InternalState = new State();
			_ServerKey = serverKey;
			_Parameters = parameters;
		}

		public SolverServerSession(RsaKey serverKey, SolverParameters parameters, State state)
			: this(serverKey, parameters)
		{
			if(state == null)
				return;
			InternalState = state;
		}


		private readonly RsaKey _ServerKey;
		public RsaKey ServerKey
		{
			get
			{
				return _ServerKey;
			}
		}

		private SolverParameters _Parameters;
		public SolverParameters Parameters
		{
			get
			{
				return _Parameters;
			}
		}

		public SolverServerStates Status
		{
			get
			{
				return InternalState.Status;
			}
		}

		public override void ConfigureEscrowedCoin(ScriptCoin escrowedCoin, Key escrowKey)
		{
			AssertState(SolverServerStates.WaitingEscrow);
			base.ConfigureEscrowedCoin(escrowedCoin, escrowKey);
			InternalState.Status = SolverServerStates.WaitingPuzzles;
		}

		public ServerCommitment[] SolvePuzzles(PuzzleValue[] puzzles)
		{
			if(puzzles == null)
				throw new ArgumentNullException(nameof(puzzles));
			if(puzzles.Length != Parameters.GetTotalCount())
				throw new ArgumentException("Expecting " + Parameters.GetTotalCount() + " puzzles");
			AssertState(SolverServerStates.WaitingPuzzles);
			List<ServerCommitment> commitments = new List<ServerCommitment>();
			List<SolvedPuzzle> solvedPuzzles = new List<SolvedPuzzle>();
			foreach(var puzzle in puzzles)
			{
				var solution = puzzle.Solve(ServerKey);
				byte[] key = null;
				var encryptedSolution = Utils.ChachaEncrypt(solution.ToBytes(), ref key);
				var solutionKey = new SolutionKey(key);
				uint160 keyHash = solutionKey.GetHash();
				commitments.Add(new ServerCommitment(keyHash, encryptedSolution));
				solvedPuzzles.Add(new SolvedPuzzle(puzzle, solutionKey, solution));
			}
			InternalState.SolvedPuzzles = solvedPuzzles.ToArray();
			InternalState.Status = SolverServerStates.WaitingRevelation;
			return commitments.ToArray();
		}

		public SolutionKey[] CheckRevelation(ClientRevelation revelation)
		{
			if(revelation == null)
				throw new ArgumentNullException($"{nameof(revelation)}");
			if(revelation.FakeIndexes.Length != Parameters.FakePuzzleCount || revelation.Solutions.Length != Parameters.FakePuzzleCount)
				throw new ArgumentException("Expecting " + Parameters.FakePuzzleCount + " puzzle solutions");
			AssertState(SolverServerStates.WaitingRevelation);



			List<SolvedPuzzle> fakePuzzles = new List<SolvedPuzzle>();
			for(int i = 0; i < Parameters.FakePuzzleCount; i++)
			{
				var index = revelation.FakeIndexes[i];
				var solvedPuzzle = InternalState.SolvedPuzzles[index];
				if(solvedPuzzle.Solution != revelation.Solutions[i])
				{
					throw new PuzzleException("Incorrect puzzle solution");
				}
				fakePuzzles.Add(solvedPuzzle);
			}

			List<SolvedPuzzle> realPuzzles = new List<SolvedPuzzle>();
			for(int i = 0; i < Parameters.GetTotalCount(); i++)
			{
				if(Array.IndexOf(revelation.FakeIndexes, i) == -1)
				{
					realPuzzles.Add(InternalState.SolvedPuzzles[i]);
				}
			}
			InternalState.SolvedPuzzles = realPuzzles.ToArray();
			InternalState.Status = SolverServerStates.WaitingBlindFactor;
			return fakePuzzles.Select(f => f.SolutionKey).ToArray();
		}

		public OfferInformation CheckBlindedFactors(BlindFactor[] blindFactors, FeeRate feeRate)
		{
			if(blindFactors == null)
				throw new ArgumentNullException(nameof(blindFactors));
			if(blindFactors.Length != Parameters.RealPuzzleCount)
				throw new ArgumentException("Expecting " + Parameters.RealPuzzleCount + " blind factors");
			AssertState(SolverServerStates.WaitingBlindFactor);
			Puzzle unblindedPuzzle = null;
			int y = 0;
			for(int i = 0; i < Parameters.RealPuzzleCount; i++)
			{
				var solvedPuzzle = InternalState.SolvedPuzzles[i];
				var unblinded = new Puzzle(Parameters.ServerKey, solvedPuzzle.Puzzle).Unblind(blindFactors[i]);
				if(unblindedPuzzle == null)
					unblindedPuzzle = unblinded;
				else if(unblinded != unblindedPuzzle)
					throw new PuzzleException("Invalid blind factor");
				y++;
			}

			InternalState.FulfillKey = new Key();

			var offerScript = GetOfferScript();

			Transaction tx = new Transaction();
			tx.AddInput(new TxIn(InternalState.EscrowedCoin.Outpoint));
			tx.AddOutput(new TxOut(InternalState.EscrowedCoin.Amount, offerScript.Hash));
			InternalState.OfferTransactionFee = feeRate.GetFee(tx.GetVirtualSize() + 72 * 2 + InternalState.EscrowedCoin.Redeem.Length);
			tx = GetUnsignedOfferTransaction();
			TransactionSignature signature = SignEscrow(tx);

			InternalState.OfferSignature = signature;
			InternalState.Status = SolverServerStates.WaitingFulfillment;
			return new OfferInformation
			{
				FulfillKey = InternalState.FulfillKey.PubKey,
				Signature = signature,
				Fee = InternalState.OfferTransactionFee
			};
		}

		private TransactionSignature SignEscrow(Transaction tx)
		{
			return tx.Inputs.AsIndexedInputs().First().Sign(InternalState.EscrowKey, InternalState.EscrowedCoin, SigHash.All);
		}

		public Transaction GetUnsignedOfferTransaction()
		{
			Transaction tx = new Transaction();
			tx.AddInput(new TxIn(InternalState.EscrowedCoin.Outpoint));
			tx.AddOutput(new TxOut(InternalState.EscrowedCoin.Amount, GetOfferScript().Hash));
			tx.Outputs[0].Value -= InternalState.OfferTransactionFee;
			return tx;
		}

		public TrustedBroadcastRequest GetSignedOfferTransaction()
		{
			AssertState(SolverServerStates.WaitingEscape);
			var offerTransaction = GetUnsignedOfferTransaction();
			TransactionBuilder txBuilder = new TransactionBuilder();
			txBuilder.Extensions.Add(new EscrowBuilderExtension());
			txBuilder.AddCoins(EscrowedCoin);
			txBuilder.AddKnownSignature(InternalState.EscrowKey.PubKey, InternalState.OfferSignature);
			txBuilder.AddKnownSignature(InternalState.GetClientEscrowPubKey(), InternalState.OfferClientSignature);
			txBuilder.SignTransactionInPlace(offerTransaction);
			return new TrustedBroadcastRequest
			{
				Key = InternalState.EscrowKey,
				Transaction = offerTransaction,
				PreviousScriptPubKey = EscrowedCoin.ScriptPubKey
			};
		}

		public SolutionKey[] GetSolutionKeys()
		{
			AssertState(SolverServerStates.WaitingEscape);
			return InternalState.SolvedPuzzles.Select(s => s.SolutionKey).ToArray();
		}

		private void AssertState(SolverServerStates state)
		{
			if(state != InternalState.Status)
				throw new InvalidOperationException("Invalid state, actual " + InternalState.Status + " while expected is " + state);
		}

		public TrustedBroadcastRequest FulfillOffer(
			TransactionSignature clientSignature,
			Script cashout,
			FeeRate feeRate)
		{
			if(clientSignature == null)
				throw new ArgumentNullException(nameof(clientSignature));
			if(feeRate == null)
				throw new ArgumentNullException(nameof(feeRate));
			AssertState(SolverServerStates.WaitingFulfillment);
			Script offerScript = GetOfferScript();

			var offer = GetUnsignedOfferTransaction();

			PubKey clientKey = AssertValidSignature(clientSignature, offer);

			TransactionBuilder builder = new TransactionBuilder();
			builder.StandardTransactionPolicy.CheckFee = false;
			builder.Extensions.Add(new EscrowBuilderExtension());
			builder.AddCoins(InternalState.EscrowedCoin);
			builder.AddKeys(InternalState.EscrowKey);
			builder.AddKnownSignature(clientKey, clientSignature);
			builder.SignTransactionInPlace(offer);
			if(!builder.Verify(offer))
				throw new PuzzleException("invalid-tumbler-signature");
			var offerCoin = offer.Outputs.AsCoins().First().ToScriptCoin(offerScript);

			var solutions = InternalState.SolvedPuzzles.Select(s => s.SolutionKey).ToArray();
			Transaction fulfill = new Transaction();
			fulfill.Inputs.Add(new TxIn(offerCoin.Outpoint));
			fulfill.Outputs.Add(new TxOut(offerCoin.Amount, cashout));

			var fulfillScript = SolverScriptBuilder.CreateFulfillScript(NBitcoin.BuilderExtensions.BuilderExtension.DummySignature, solutions);
			fulfill.Inputs[0].ScriptSig = fulfillScript + Op.GetPushOp(offerCoin.Redeem.ToBytes());

			fulfill.Outputs[0].Value -= feeRate.GetFee(fulfill.GetVirtualSize());

			var signature = fulfill.Inputs.AsIndexedInputs().First().Sign(InternalState.FulfillKey, offerCoin, SigHash.All);
			fulfillScript = SolverScriptBuilder.CreateFulfillScript(signature, solutions);
			fulfill.Inputs[0].ScriptSig = fulfillScript + Op.GetPushOp(offerCoin.Redeem.ToBytes());

			InternalState.OfferClientSignature = clientSignature;
			InternalState.Status = SolverServerStates.WaitingEscape;
			return new TrustedBroadcastRequest
			{
				Key = InternalState.FulfillKey,
				PreviousScriptPubKey = offerCoin.ScriptPubKey,
				Transaction = fulfill
			};
		}

		private PubKey AssertValidSignature(TransactionSignature clientSignature, Transaction offer)
		{
			var signedHash = offer.Inputs.AsIndexedInputs().First().GetSignatureHash(InternalState.EscrowedCoin, clientSignature.SigHash);
			var clientKey = InternalState.GetClientEscrowPubKey();
			if(!clientKey.Verify(signedHash, clientSignature.Signature))
				throw new PuzzleException("invalid-client-signature");
			return clientKey;
		}

		public Transaction GetSignedEscapeTransaction(TransactionSignature clientSignature, Script cashout)
		{
			AssertState(SolverServerStates.WaitingEscape);

			var escapeTx = GetUnsignedOfferTransaction();
			escapeTx.Outputs[0].ScriptPubKey = cashout;
			var clientKey = AssertValidSignature(clientSignature, escapeTx);

			TransactionBuilder builder = new TransactionBuilder();
			builder.StandardTransactionPolicy.CheckFee = false;
			builder.Extensions.Add(new EscrowBuilderExtension());
			builder.AddCoins(InternalState.EscrowedCoin);		
			builder.AddKnownSignature(clientKey, clientSignature);
			//This add the known signature if correct SigHash
			builder.SignTransactionInPlace(escapeTx, SigHash.None | SigHash.AnyoneCanPay);

			//This sign SigHash.All
			builder.AddKeys(InternalState.EscrowKey);
			builder.SignTransactionInPlace(escapeTx);
			if(!builder.Verify(escapeTx))
				throw new PuzzleException("invalid-tumbler-signature");

			return escapeTx;
		}

		private Script GetOfferScript()
		{
			var escrow = EscrowScriptBuilder.ExtractEscrowScriptPubKeyParameters(InternalState.EscrowedCoin.Redeem);
			return SolverScriptBuilder.CreateOfferScript(new OfferScriptPubKeyParameters
			{
				Hashes = InternalState.SolvedPuzzles.Select(p => p.SolutionKey.GetHash()).ToArray(),
				FulfillKey = InternalState.FulfillKey.PubKey,
				RedeemKey = escrow.RedeemKey,
				Expiration = escrow.LockTime
			});
		}

		public Script GetOfferScriptPubKey()
		{
			return GetOfferScript().Hash.ScriptPubKey;
		}

		public override LockTime GetLockTime(CycleParameters cycle)
		{
			return cycle.GetClientLockTime();
		}
	}
}
